import inspect
import os
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional, TextIO, Union

from _finetuner.runner.stubs import model as model_stub
from docarray import DocumentArray

from finetuner.constants import (
    DEFAULT_FINETUNER_HOST,
    DEFAULT_HUBBLE_REGISTRY,
    HOST,
    HUBBLE_REGISTRY,
)
from finetuner.data import CSVOptions
from finetuner.run import Run
from hubble import login_required

if HOST not in os.environ:
    os.environ[HOST] = DEFAULT_FINETUNER_HOST

if HUBBLE_REGISTRY not in os.environ:
    os.environ[HUBBLE_REGISTRY] = DEFAULT_HUBBLE_REGISTRY

from finetuner import callback, model
from finetuner.console import print_model_table
from finetuner.data import build_encoding_dataset
from finetuner.experiment import Experiment
from finetuner.finetuner import Finetuner
from finetuner.model import list_model_classes

if TYPE_CHECKING:
    import numpy as np
    from _finetuner.models.inference import InferenceEngine

ft = Finetuner()


def login(force: bool = False, interactive: Optional[bool] = None):
    """
    Login to Jina AI Cloud to use cloud-based fine-tuning.
    Thereby, an authentication token is
    generated which can be read with the :func:`~finetuner.get_token` function.

    :param force: If set to true, an existing token will be overwritten. Otherwise,
        you will not login again, if a valid token already exists.
    :param interactive: Interactive mode should be set in Jupyter environments.
    """
    ft.login(force=force, interactive=interactive)


def list_callbacks() -> Dict[str, callback.CallbackStubType]:
    """List available callbacks."""
    return {
        name: obj for name, obj in inspect.getmembers(callback) if inspect.isclass(obj)
    }


def _build_name_stub_map() -> Dict[str, model_stub.ModelStubType]:
    rv = {}
    members = inspect.getmembers(model_stub, inspect.isclass)
    for name, stub in members:
        if name != 'MLPStub' and not name.startswith('_') and type(stub) != type:
            rv[stub.name] = stub
    return rv


def list_models() -> List[str]:
    """List available models."""
    return [name for name in list_model_classes()]


def list_model_options() -> Dict[str, List[Dict[str, Any]]]:
    """List available options per model."""
    return {
        name: [
            {
                'name': parameter.name,
                'type': parameter.annotation,
                'required': parameter.default == parameter.empty,
                'default': (
                    parameter.default if parameter.default != parameter.empty else None
                ),
            }
            for parameter in inspect.signature(
                _model_class.__init__
            ).parameters.values()
            if parameter.name != 'self'
        ]
        for name, _model_class in list_model_classes().items()
    }


def describe_models(task: Optional[str] = None) -> None:
    """Print model information, such as name, task, output dimension, architecture
    and description as a table.

    :param task: The task for the backbone model, one of `text-to-text`,
        `text-to-image`, `image-to-image`. If not provided, will print all backbone
        models.

    """
    print_model_table(model, task=task)


@login_required
def fit(
    model: str,
    train_data: Union[str, TextIO, DocumentArray],
    eval_data: Optional[Union[str, TextIO, DocumentArray]] = None,
    val_split: float = 0.0,
    model_artifact: Optional[str] = None,
    run_name: Optional[str] = None,
    description: Optional[str] = None,
    experiment_name: Optional[str] = None,
    model_options: Optional[Dict[str, Any]] = None,
    loss: str = 'TripletMarginLoss',
    miner: Optional[str] = None,
    miner_options: Optional[Dict[str, Any]] = None,
    optimizer: str = 'Adam',
    optimizer_options: Optional[Dict[str, Any]] = None,
    learning_rate: Optional[float] = None,
    epochs: int = 5,
    batch_size: Optional[int] = None,
    callbacks: Optional[List[callback.CallbackStubType]] = None,
    scheduler: Optional[str] = None,
    scheduler_options: Optional[Dict[str, Any]] = None,
    freeze: bool = False,
    output_dim: Optional[int] = None,
    device: str = 'cuda',
    num_workers: int = 4,
    to_onnx: bool = False,
    csv_options: Optional[CSVOptions] = None,
    public: bool = False,
    num_items_per_class: int = 4,
    sampler: str = 'auto',
    loss_optimizer: Optional[str] = None,
    loss_optimizer_options: Optional[Dict[str, Any]] = None,
) -> Run:
    """Create a Finetuner :class:`Run`, calling this function will submit a fine-tuning
    job to the Jina AI Cloud.

    :param model: The name of model to be fine-tuned. Run `finetuner.list_models()` or
        `finetuner.describe_models()` to see the available model names.
    :param train_data: Either a `DocumentArray` for training data, a name of the
        `DocumentArray` that is pushed on Jina AI Cloud or a path to a CSV file.
    :param eval_data: Either a `DocumentArray` for evaluation data, a name of the
        `DocumentArray` that is pushed on Jina AI Cloud or a path to a CSV file.
    :param val_split: Determines which portion of the `train_data` is held out
        for calculating a validation loss. If it is set to 0, or an `eval_data`
        parameter is provided, no data is held out from the training data. Instead, the
        `eval_data` is used to calculate the validation loss if it is provided.
    :param model_artifact: To continue training the training of a model which was
        fine-tuned by a previous run, you can provide the artifact id of this model,
        which you can get via :meth:`Run.artifact_id`.
    :param run_name: Name of the run.
    :param description: Run description.
    :param experiment_name: Name of the experiment.
    :param model_options: Additional arguments to pass to the model construction. These
        are model specific options and are different depending on the model you choose.
        Run `finetuner.list_model_options()` to see available options for every model.
    :param loss: Name of the loss function used for fine-tuning. Default is
        `TripletMarginLoss`. Options: `CosFaceLoss`, `NTXLoss`, `AngularLoss`,
        `ArcFaceLoss`, `BaseMetricLossFunction`, `MultipleLosses`,
        `CentroidTripletLoss`, `CircleLoss`, `ContrastiveLoss`, `CrossBatchMemory`,
        `FastAPLoss`, `GenericPairLoss`, `IntraPairVarianceLoss`,
        `LargeMarginSoftmaxLoss`, `GeneralizedLiftedStructureLoss`,
        `LiftedStructureLoss`, `MarginLoss`, `EmbeddingRegularizerMixin`,
        `WeightRegularizerMixin`, `MultiSimilarityLoss`, `NPairsLoss`, `NCALoss`,
        `NormalizedSoftmaxLoss`, `ProxyAnchorLoss`, `ProxyNCALoss`,
        `SignalToNoiseRatioContrastiveLoss`, `SoftTripleLoss`, `SphereFaceLoss`,
        `SupConLoss`, `TripletMarginLoss`, `TupletMarginLoss`, `VICRegLoss`,
        `CLIPLoss`.
    :param miner: Name of the miner to create tuple indices for the loss function.
        Options: `AngularMiner`, `BaseMiner`, `BaseSubsetBatchMiner`, `BaseTupleMiner`,
        `BatchEasyHardMiner`, `BatchHardMiner`, `DistanceWeightedMiner`, `HDCMiner`,
        `EmbeddingsAlreadyPackagedAsTriplets`, `MaximumLossMiner`, `PairMarginMiner`,
        `MultiSimilarityMiner`, `TripletMarginMiner`, `UniformHistogramMiner`.
    :param miner_options: Additional parameters to pass to the miner construction. The
        set of applicable parameters is specific to the miner you choose. Details on
        the parameters can be found in the `PyTorch Metric Learning documentation
        <https://kevinmusgrave.github.io/pytorch-metric-learning/miners/>`_
    :param optimizer: Name of the optimizer used for fine-tuning. Options: `Adadelta`,
        `Adagrad`, `Adam`, `AdamW`, `SparseAdam`, `Adamax`, `ASGD`, `LBFGS`, `NAdam`,
        `RAdam`, `RMSprop`, `Rprop`, `SGD`.
    :param optimizer_options: Additional parameters to pass to the optimizer
        construction. The set of applicable parameters is specific to the optimizer you
        choose. Details on the parameters can be found in the `PyTorch documentation
        <https://pytorch.org/docs/stable/optim.html>`_
    :param learning_rate: learning rate for the optimizer.
    :param epochs: Number of epochs for fine-tuning.
    :param batch_size: Number of items to include in a batch. If not set, the
        batch size will be configured automatically.
    :param callbacks: List of callback stub objects.
        subpackage for available options, or run `finetuner.list_callbacks()`.
    :param scheduler: Name of a scheduler to use for learning rate scheduling.
        Supported types are: `linear`, `cosine`, `cosine_with_restarts`, `polynomial`,
        `constant`, `constant_with_warmup`.
    :param scheduler_options: Dictionary of additional parameters to pass to the
        scheduler: `num_warmup_steps`, `num_training_steps`, and `scheduler_step`
        (either `batch` or `epoch`).
    :param freeze: If set to `True`, will freeze all layers except the last one.
    :param output_dim: The expected output dimension as `int`.
        If set, will attach a projection head.
    :param device: Whether to use the CPU, if set to `cuda`, a Nvidia GPU will be used.
        otherwise use `cpu` to run a cpu job.
    :param num_workers: Number of CPU workers. If `cpu: False` this is the number of
        workers used by the dataloader.
    :param to_onnx: Set this parameter as `True` to convert the model to an onnx model.
        Please note that not all models support this. If this parameter is set, please
        pass `is_onnx` when making inference, e.g., when calling the `get_model`
        function.
    :param csv_options: A :class:`CSVOptions` object containing options used for
        reading in training and evaluation data from a CSV file, if they are
        provided as such.
    :param public: A boolean value indicates if the artifact is public. It should be
        set to `True` if you would like to share your fine-tuned model with others.
    :param num_items_per_class: How many items per class (unique labels) to include
        in a batch. For example, if ``batch_size`` is 20, and
        ``num_items_per_class`` is 4, the batch will consist of 4 items for each of
        the 5 classes. Batch size must be divisible by `num_items_per_class`.
    :param sampler: Determines which sampling method will be used if the
        data is labeled. Default is `auto`, meaning that the sampler,
        will be the default for the loss function used.
        Setting to `class` will result in the `ClassSampler` being used, and setting to
        `random` will result in the `RandomSampler` being used.
        If set to `random` then `num_items_per_class` is not used.
    :param loss_optimizer: Name of the optimizer used for fine-tuning to loss function,
        if it is a function that requires an optimizer. Options: `Adadelta`, `Adagrad`,
        `Adam`, `AdamW`, `SparseAdam`, `Adamax`, `ASGD`, `LBFGS`, `NAdam`, `RAdam`,
        `RMSprop`, `Rprop`, `SGD`. If left as None then optimizer specified by
        the `optimizer` argument will be used instead.
    :param loss_optimizer_options: Additional parameters to pass to the optimizer of
        the loss function. The set of applicable parameters is specific to the optimizer
        you choose. Details on the parameters can be found in the `PyTorch documentation
        <https://pytorch.org/docs/stable/optim.html>`_.

    .. note::
       Unless necessary, please stick with `device="cuda"`, `cpu` training could be
       extremely slow and inefficient.
    """

    return ft.create_run(
        model=model,
        train_data=train_data,
        eval_data=eval_data,
        val_split=val_split,
        model_artifact=model_artifact,
        run_name=run_name,
        description=description,
        experiment_name=experiment_name,
        model_options=model_options,
        loss=loss,
        miner=miner,
        miner_options=miner_options,
        optimizer=optimizer,
        optimizer_options=optimizer_options,
        learning_rate=learning_rate,
        epochs=epochs,
        batch_size=batch_size,
        callbacks=callbacks,
        scheduler=scheduler,
        scheduler_options=scheduler_options,
        freeze=freeze,
        output_dim=output_dim,
        device=device,
        num_workers=num_workers,
        to_onnx=to_onnx,
        csv_options=csv_options,
        public=public,
        num_items_per_class=num_items_per_class,
        sampler=sampler,
        loss_optimizer=loss_optimizer,
        loss_optimizer_options=loss_optimizer_options,
    )


# `create_run` and `fit` do the same
create_run = fit


def get_run(run_name: str, experiment_name: Optional[str] = None) -> Run:
    """Get a :class:`Run` by its name and (optional) :class:`Experiment` name.

    If an experiment name is not specified, we'll look for the run in the default
    experiment.

    :param run_name: Name of the :class:`Run`.
    :param experiment_name: Optional name of the :class:`Experiment`.
    :return: A :class:`Run` object.
    """
    return ft.get_run(run_name=run_name, experiment_name=experiment_name)


def list_runs(
    experiment_name: Optional[str] = None, page: int = 1, size: int = 50
) -> List[Run]:
    """List all created :class:`Run` inside a given :class:`Experiment`.

    If no :class:`Experiment` is specified, list :class:`Run` for all available
      :class:`Experiment`.
    :param experiment_name: The name of the :class:`Experiment`.
    :param page: The page index.
    :param size: Number of :class:`Run` to retrieve.
    :return: List of all :class:`Run`.

    ..note:: `page` and `size` works together. For example, page 1 size 50 gives
        the 50 runs in the first page. To get 50-100, set `page` as 2.
    ..note:: The maximum number for `size` per page is 100.
    """
    return ft.list_runs(experiment_name=experiment_name, page=page, size=size)


def delete_run(run_name: str, experiment_name: Optional[str] = None) -> None:
    """Delete a :class:`Run` given a `run_name` and
        optional `experiment_name`.

    If an experiment name is not specified, we'll look for the run in the default
    experiment.

    :param run_name: Name of the run. View your runs with `list_runs`.
    :param experiment_name: Optional name of the experiment.
    """
    ft.delete_run(run_name=run_name, experiment_name=experiment_name)


def delete_runs(experiment_name: Optional[str] = None) -> None:
    """Delete all :class:`Run` given an optional `experiment_name`.

    If an experiment name is not specified, we'll delete every run across all
    experiments.

    :param experiment_name: Optional name of the experiment.
        View your experiment names with `list_experiments()`.
    """
    ft.delete_runs(experiment_name=experiment_name)


def create_experiment(name: str = 'default') -> Experiment:
    """Create an :class:`Experiment`.

    :param name: The name of the experiment. If not provided,
        the experiment is named as `default`.
    :return: An `Experiment` object.
    """
    return ft.create_experiment(name=name)


def get_experiment(name: str) -> Experiment:
    """Get an :class:`Experiment` given a `name`.

    :param name: Name of the experiment.
    :return: An `Experiment` object.
    """
    return ft.get_experiment(name=name)


def list_experiments(page: int = 1, size: int = 50) -> List[Experiment]:
    """List all :class:`Experiment`.

    :param page: The page index.
    :param size: The number of experiments to retrieve.
    :return: A list of :class:`Experiment` instance.

    ..note:: `page` and `size` works together. For example, page 1 size 50 gives
        the 50 experiments in the first page. To get 50-100, set `page` as 2.
    ..note:: The maximum number for `size` per page is 100.
    """
    return ft.list_experiments(page=page, size=size)


def delete_experiment(name: str) -> Experiment:
    """Delete an :class:`Experiment` given a `name`.

    :param name: Name of the experiment.
        View your experiment names with `list_experiments()`.
    :return: Deleted experiment.
    """
    return ft.delete_experiment(name=name)


def delete_experiments() -> List[Experiment]:
    """Delete all :class:`Experiment`.
    :return: List of deleted experiments.
    """
    return ft.delete_experiments()


def get_token() -> str:
    """Get user token from the Jina AI Cloud, :meth:`login` is required.

    :return: user token as string object.
    """
    return ft.get_token()


def build_model(
    name: str,
    model_options: Optional[Dict[str, Any]] = None,
    batch_size: int = 32,
    select_model: Optional[str] = None,
    device: Optional[str] = None,
    is_onnx: bool = False,
) -> 'InferenceEngine':
    """
    Builds a pre-trained model given a `name`.

    :param name: Refers to a pre-trained model, see
        https://finetuner.jina.ai/walkthrough/choose-backbone/  or use the
        :meth:`finetuner.describe_models()` function for a list of all
        supported models.
    :param model_options: A dictionary of model specific options.
    :param batch_size: Incoming documents are fed to the graph in batches, both to
        speed-up inference and avoid memory errors. This argument controls the
        number of documents that will be put in each batch.
    :param select_model: Finetuner run artifacts might contain multiple models. In
        such cases you can select which model to deploy using this argument. For CLIP
        fine-tuning, you can choose either `clip-vision` or `clip-text`.
    :param device: Whether to use the CPU, if set to `cuda`, a Nvidia GPU will be used.
        otherwise use `cpu` to run a cpu job.
    :param is_onnx: The model output format, either `onnx` or `pt`.
    :return: an instance of :class:'TorchInferenceEngine' or
        :class:`ONNXINferenceEngine`.
    """
    import torch
    from _finetuner.models.inference import (
        ONNXRuntimeInferenceEngine,
        TorchInferenceEngine,
    )
    from _finetuner.runner.model import EmbeddingModel

    if not device:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

    stub = model_stub.get_stub(
        name, select_model=select_model, model_options=model_options or {}
    )

    model = EmbeddingModel(stub=stub, download_pretrained=True)
    if not is_onnx:
        return TorchInferenceEngine(
            artifact=model,
            batch_size=batch_size,
            device=device,
            download_pretrained=True,
        )
    else:
        return ONNXRuntimeInferenceEngine(
            artifact=model,
            batch_size=batch_size,
            device=device,
        )


def get_model(
    artifact: str,
    token: Optional[str] = None,
    batch_size: int = 32,
    select_model: Optional[str] = None,
    device: Optional[str] = None,
    logging_level: str = 'WARNING',
    is_onnx: bool = False,
) -> 'InferenceEngine':
    """Re-build the model based on the model inference session with ONNX.

    :param artifact: Specify a finetuner run artifact. Can be a path to a local
        directory, a path to a local zip file, or a Hubble artifact ID. Individual
        model artifacts (model sub-folders inside the run artifacts) can also be
        specified using this argument.
    :param token: A Jina authentication token required for pulling artifacts from
        Hubble. If not provided, the Hubble client will try to find one either in a
        local cache folder or in the environment.
    :param batch_size: Incoming documents are fed to the graph in batches, both to
        speed-up inference and avoid memory errors. This argument controls the
        number of documents that will be put in each batch.
    :param select_model: Finetuner run artifacts might contain multiple models. In
        such cases you can select which model to deploy using this argument. For CLIP
        fine-tuning, you can choose either `clip-vision` or `clip-text`.
    :param device: Whether to use the CPU, if set to `cuda`, a Nvidia GPU will be used.
        otherwise use `cpu` to run a cpu job.
    :param logging_level: The executor logging level. See
        https://docs.python.org/3/library/logging.html#logging-levels for available
        options.
    :param is_onnx: The model output format, either `onnx` or `pt`.
    :returns: An instance of :class:`ONNXRuntimeInferenceEngine`.

    ..Note::
      please install finetuner[full] to include all the dependencies.
    """

    import torch
    from _finetuner.models.inference import (
        ONNXRuntimeInferenceEngine,
        TorchInferenceEngine,
    )

    if not device:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

    if device == 'cuda' and is_onnx:
        warnings.warn(
            message='You are using cuda device for ONNX inference, please consider'
            'calling `pip install onnxruntime-gpu` to speed up inference.',
            category=RuntimeWarning,
        )
    if is_onnx:
        inference_engine = ONNXRuntimeInferenceEngine(
            artifact=artifact,
            token=token,
            batch_size=batch_size,
            select_model=select_model,
            device=device,
            logging_level=logging_level,
        )
    else:
        inference_engine = TorchInferenceEngine(
            artifact=artifact,
            token=token,
            batch_size=batch_size,
            select_model=select_model,
            device=device,
            logging_level=logging_level,
            download_pretrained=False,
        )
    return inference_engine


def encode(
    model: 'InferenceEngine',
    data: Union[DocumentArray, List[str]],
    batch_size: int = 32,
) -> Union[DocumentArray, 'np.ndarray']:
    """Preprocess, collate and encode the `list or :class:`DocumentArray`
        with embeddings.

    :param model: The model to be used to encode `DocumentArray`. In this case
        an instance of `ONNXRuntimeInferenceEngine` or `TorchInferenceEngine`
        produced by :meth:`finetuner.get_model()`
    :param data: The `DocumentArray` object to be encoded.
    :param batch_size: Incoming documents are fed to the graph in batches, both to
        speed-up inference and avoid memory errors. This argument controls the
        number of documents that will be put in each batch.
    :returns: `DocumentArray` filled with embeddings.

    ..Note::
      please install "finetuner[full]" to include all the dependencies.
    """

    from _finetuner.models.inference import ONNXRuntimeInferenceEngine

    if isinstance(data, DocumentArray):
        return_da = True
    else:
        data = build_encoding_dataset(model=model, data=data)
        return_da = False

    for batch in data.batch(batch_size, show_progress=True):
        if isinstance(model, ONNXRuntimeInferenceEngine):
            inputs = model._run_data_pipeline(batch)
            inputs = model._flatten_inputs(inputs)
            model._check_input_names(inputs)
            output_shape = model._infer_output_shape(inputs)
            inputs = model._move_to_device(inputs)
            output = model.run(inputs, output_shape)
            batch.embeddings = output.cpu().numpy()
        else:
            inputs = model._run_data_pipeline(batch)
            inputs = model._flatten_inputs(inputs)
            model._check_input_names(inputs)
            inputs = model._move_to_device(inputs)
            output = model.run(inputs)
            batch.embeddings = output.detach().cpu().numpy()

    return data if return_da else data.embeddings



import abc
import csv
import os
from abc import ABC
from contextlib import nullcontext
from dataclasses import dataclass
from io import StringIO
from typing import TYPE_CHECKING, List, Optional, TextIO, Tuple, Union

from _finetuner.runner.stubs.model import get_stub
from docarray import Document, DocumentArray
from docarray.document.generators import _subsample
from docarray.document.mixins.helper import _is_uri
from genericpath import isfile

from finetuner.constants import DEFAULT_TAG_KEY, DEFAULT_TAG_SCORE_KEY

if TYPE_CHECKING:
    from _finetuner.models.inference import InferenceEngine


@dataclass
class CSVOptions:
    """Class containing options for reading CSV files

    :param size: The number of rows that will be sampled.
    :param sampling_rate: The sampling rate between [0, 1] indicating how many lines of
        the CSV are skipped. a sampling rate of 1 means that none are skipped, 0.5
        means that half are skipped, and 0 means that all lines are skipped.
    :param dialect: A description of the expected format of the CSV, can either be an
        object of the :class:`csv.Dialect` class, or one of the strings returned by the
        :meth:`csv.list_dialects()' function.
    :param encoding: The encoding of the CSV file.
    :param is_labeled: Whether the second column of the CSV represents a label that
        should be assigned to the item in the first column (True), or if it is another
        item that should be semantically close to the first (False).
    :param convert_to_blob: Whether uris to local files should be converted to blobs
    :param create_point_clouds: Determines whether from uris to local 3D mesh files
        should point clouds be sampled.
    :param point_cloud_size: Determines the number of points sampled from a mesh to
        create a point cloud.
    """

    size: Optional[int] = None
    sampling_rate: Optional[float] = None
    dialect: Union[str, 'csv.Dialect'] = 'auto'
    encoding: str = 'utf-8'
    is_labeled: bool = False
    convert_to_blob: bool = True
    create_point_clouds: bool = True
    point_cloud_size: int = 2048


class _CSVParser(ABC):
    def __init__(
        self,
        file: Union[str, TextIO, StringIO],
        task: str,
        options: Optional[CSVOptions] = None,
    ):
        self._file = file
        self._task = task
        self._options = options or CSVOptions()
        if isinstance(self._options.dialect, str) and self._options.dialect == 'auto':
            self._dialect, _ = get_csv_file_dialect_columns(
                self._file, encoding=options.encoding
            )
            self._options.dialect = self._dialect
        self._file_ctx = get_csv_file_context(file=file, encoding=options.encoding)

    @abc.abstractmethod
    def parse(self):
        ...


class LabeledCSVParser(_CSVParser):
    """
    CSV has two columns where the first column is the data, the second column is the
    label. To use the handler, make sure csv contains two columns and `is_labeled=True`.
    """

    def __init__(
        self,
        file: Union[str, TextIO, StringIO],
        task: str,
        options: Optional[CSVOptions] = None,
    ):
        super().__init__(file, task, options)

    def parse(self):
        with self._file_ctx as fp:

            lines = csv.reader(fp, dialect=self._options.dialect)

            for columns in _subsample(
                lines, self._options.size, self._options.sampling_rate
            ):
                col1, col2 = columns
                modality_col1, modality_col2 = check_columns(self._task, col1, col2)
                doc = create_document(
                    modality_col1,
                    col1,
                    self._options.convert_to_blob,
                    self._options.create_point_clouds,
                    point_cloud_size=self._options.point_cloud_size,
                )
                doc.tags[DEFAULT_TAG_KEY] = col2
                yield doc


class QueryDocumentRelationsParser(_CSVParser):
    """
    In the case that user do not have explicitly annotated labels,
    but rather a set of query-document pairs which express that a document is
    relevant to a query, or form as a text-image pair.
    """

    def __init__(
        self,
        file: Union[str, TextIO, StringIO],
        task: str,
        options: Optional[CSVOptions] = None,
    ):
        super().__init__(file, task, options)

    def parse(self):
        with self._file_ctx as fp:

            queries = {}
            artificial_label = 0
            modality_col1, modality_col2 = None, None
            lines = csv.reader(fp, dialect=self._options.dialect)

            for columns in _subsample(
                lines, self._options.size, self._options.sampling_rate
            ):
                col1, col2 = columns
                if col1 in queries and col2 in queries:
                    continue
                if not modality_col1:
                    modality_col1, modality_col2 = check_columns(self._task, col1, col2)
                doc1 = create_document(
                    modality_col1,
                    col1,
                    self._options.convert_to_blob,
                    self._options.create_point_clouds,
                    point_cloud_size=self._options.point_cloud_size,
                )
                doc2 = create_document(
                    modality_col2,
                    col2,
                    self._options.convert_to_blob,
                    self._options.create_point_clouds,
                    point_cloud_size=self._options.point_cloud_size,
                )
                if col1 in queries:
                    queries[col2] = queries[col1]
                    doc2.tags[DEFAULT_TAG_KEY] = queries[col1]
                    # only yield d2
                else:
                    queries[col1] = artificial_label
                    queries[col2] = artificial_label
                    # yield both
                    doc1.tags[DEFAULT_TAG_KEY] = queries[col1]
                    doc2.tags[DEFAULT_TAG_KEY] = queries[col1]
                    artificial_label += 1

                if modality_col1 == modality_col2:
                    doc1.modality = modality_col1
                    doc2.modality = modality_col1
                    if DEFAULT_TAG_KEY in doc1.tags:
                        yield doc1
                    if DEFAULT_TAG_KEY in doc2.tags:
                        yield doc2
                else:
                    # different modalities, for CLIP
                    doc1.modality = modality_col1
                    doc2.modality = modality_col2
                    yield Document(
                        chunks=[doc1, doc2], tags={DEFAULT_TAG_KEY: queries[col1]}
                    )


class PairwiseScoreParser(_CSVParser):
    """
    CSV has three columns, column1, column2 and a float value indicates the
    similarity between column1 and column2.
    """

    def __init__(
        self,
        file: Union[str, TextIO, StringIO],
        task: str,
        options: Optional[CSVOptions] = None,
    ):
        super().__init__(file, task, options)

    def parse(self):
        with self._file_ctx as fp:

            lines = csv.reader(fp, dialect=self._options.dialect)

            for columns in _subsample(
                lines, self._options.size, self._options.sampling_rate
            ):
                col1, col2, col3 = columns
                modality_col1, modality_col2 = check_columns(self._task, col1, col2)
                doc1 = create_document(
                    modality_col1,
                    col1,
                    self._options.convert_to_blob,
                    self._options.create_point_clouds,
                    point_cloud_size=self._options.point_cloud_size,
                )
                doc2 = create_document(
                    modality_col2,
                    col2,
                    self._options.convert_to_blob,
                    self._options.create_point_clouds,
                    point_cloud_size=self._options.point_cloud_size,
                )
                yield Document(chunks=[doc1, doc2], tags={DEFAULT_TAG_SCORE_KEY: col3})


class CSVContext:
    """
    A CSV context switch class with conditions to parse CSVs into DocumentArray.

    :param model: The model being used, to get model stub and associated task.
    :param options: an instance of :class`CSVOptions`.
    """

    def __init__(
        self,
        model: str,
        options: Optional[CSVOptions] = None,
    ):
        self._model = model
        self._options = options or CSVOptions()
        if model == 'mlp':
            self._task = 'image-to-image'
        else:
            model_stub = get_stub(
                model,
                select_model='clip-text',
            )
            # for clip select_model is mandatory, though any model will get us the task
            self._task = model_stub.task

    def _get_csv_parser(self, data: Union[str, TextIO]):
        if self._options.is_labeled:
            return LabeledCSVParser(file=data, task=self._task, options=self._options)
        else:
            _, num_columns = get_csv_file_dialect_columns(
                file=data, encoding=self._options.encoding
            )
            if num_columns == 2:
                return QueryDocumentRelationsParser(
                    file=data, task=self._task, options=self._options
                )
            elif num_columns == 3:
                return PairwiseScoreParser(
                    file=data, task=self._task, options=self._options
                )
            else:
                raise TypeError('Can not determine the context of the csv.')

    def build_dataset(self, data: Union[str, TextIO, StringIO, DocumentArray]):
        if (
            isinstance(data, TextIO)
            or isinstance(data, StringIO)
            or (isinstance(data, str) and isfile(data))
        ):
            parser = self._get_csv_parser(data=data)
            da_generator = parser.parse()
            data = DocumentArray(da_generator)

        return data


def get_csv_file_context(file: Union[str, TextIO, StringIO], encoding: str):
    """Get csv file context, such as `file_ctx`, csv dialect and number of columns."""
    if hasattr(file, 'read'):
        return nullcontext(file)
    return open(file, 'r', encoding=encoding)


def get_csv_file_dialect_columns(file: str, encoding: str):
    """Get csv dialect and number of columns of the csv."""
    file_ctx = get_csv_file_context(file=file, encoding=encoding)
    with file_ctx as fp:
        try:
            dialect = csv.Sniffer().sniff(fp.read(1024))
            fp.seek(0)
        except Exception:
            dialect = 'excel'  #: can not sniff delimiter, use default dialect
        try:
            reader = csv.reader(fp, dialect=dialect)
            num_columns = len(next(reader))
        except StopIteration:
            raise IOError('CSV file not exist or is empty')
        fp.seek(0)
        return dialect, num_columns


def build_encoding_dataset(model: 'InferenceEngine', data: List[str]) -> DocumentArray:
    """If data has been provided as a list, a :class:`DocumentArray` is created
    from the elements of the list
    """
    modalities = model._metadata['preprocess_types']
    if model._select_model:
        if model._select_model == 'clip-text':
            task = 'text'
        else:
            task = 'image'
    elif list(modalities)[0] == ['features']:
        raise ValueError('MLP model does not support values from a list.')
    else:
        task = list(modalities)[0]

    data = DocumentArray(
        [Document(text=d) if task == 'text' else Document(uri=d) for d in data]
    )

    return data


def check_columns(
    task: str,
    col1: str,
    col2: str,
) -> Tuple[str, str]:
    """Determines the expected modalities of each column using the task argument,
        Then checks the given row of the CSV to confirm that it contains valid data

    :param task: The task of the model being used.
    :param col1: A single value from the first column of the CSV.
    :param col2: A single value from the second column of the CSV.
    :return: The expected modality of each column
    """

    if task == 'any':
        raise ValueError('MLP model does not support values read in from CSV files.')

    if len(task.split('-to-')) == 2:
        modality_col1, modality_col2 = task.split('-to-')
    else:
        raise ValueError(f'Model has invalid task: {task}')

    if modality_col1 == 'text' and modality_col2 == 'image':
        if _is_uri(col1) and not _is_uri(col2):
            modality_col1 = 'image'
            modality_col2 = 'text'
        elif not _is_uri(col2):
            raise ValueError(
                (
                    'Uri required in at least one colum '
                    'for model with task: '
                    'text-to-image'
                    '.'
                )
            )
    return modality_col1, modality_col2


def create_document(
    modality: str,
    column: str,
    convert_to_blob: bool,
    create_point_clouds: bool,
    point_cloud_size: int = 2048,
) -> Document:
    """Checks the expected modality of the value in the given column
        and creates a :class:`Document` with that value

    :param modality: The expected modality of the value in the given column
    :param column: A single value of a column
    :param convert_to_blob: Whether uris to local image files should be converted to
        blobs.
    :param create_point_clouds: Whether from uris to local 3D mesh files should point
        clouds be sampled.
    :param point_cloud_size: Determines the number of points sampled from a mesh to
        create a point cloud.
    """
    if modality == 'image':
        if _is_uri(column):
            doc = Document(uri=column)
            if convert_to_blob and os.path.isfile(column):
                doc.load_uri_to_blob()
        else:
            raise ValueError(f'Expected uri in column 1, got {column}')
    elif modality == 'mesh':
        if _is_uri(column):
            doc = Document(uri=column)
            if create_point_clouds and os.path.isfile(column):
                doc.load_uri_to_point_cloud_tensor(point_cloud_size)
        else:
            raise ValueError(f'Expected uri in column 1, got {column}')
    else:
        doc = Document(content=column)

    return doc



import random

# moby-like silly name generator
# taken from
# https://github.com/moby/moby/blob/master/pkg/namesgenerator/names-generator.go


adjectives = [
    'admiring',
    'adoring',
    'affectionate',
    'agitated',
    'amazing',
    'angry',
    'awesome',
    'beautiful',
    'blissful',
    'bold',
    'boring',
    'brave',
    'busy',
    'charming',
    'clever',
    'cool',
    'compassionate',
    'competent',
    'condescending',
    'confident',
    'cranky',
    'crazy',
    'dazzling',
    'determined',
    'distracted',
    'dreamy',
    'eager',
    'ecstatic',
    'elastic',
    'elated',
    'elegant',
    'eloquent',
    'epic',
    'exciting',
    'fervent',
    'festive',
    'flamboyant',
    'focused',
    'friendly',
    'frosty',
    'funny',
    'gallant',
    'gifted',
    'goofy',
    'gracious',
    'great',
    'happy',
    'hardcore',
    'heuristic',
    'hopeful',
    'hungry',
    'infallible',
    'inspiring',
    'interesting',
    'intelligent',
    'jolly',
    'jovial',
    'keen',
    'kind',
    'laughing',
    'loving',
    'lucid',
    'magical',
    'mystifying',
    'modest',
    'musing',
    'naughty',
    'nervous',
    'nice',
    'nifty',
    'nostalgic',
    'objective',
    'optimistic',
    'peaceful',
    'pedantic',
    'pensive',
    'practical',
    'priceless',
    'quirky',
    'quizzical',
    'recursing',
    'relaxed',
    'reverent',
    'romantic',
    'sad',
    'serene',
    'sharp',
    'silly',
    'sleepy',
    'stoic',
    'strange',
    'stupefied',
    'suspicious',
    'sweet',
    'tender',
    'thirsty',
    'trusting',
    'unruffled',
    'upbeat',
    'vibrant',
    'vigilant',
    'vigorous',
    'wizardly',
    'wonderful',
    'xenodochial',
    'youthful',
    'zealous',
    'zen',
]


surnames = [
    'albattani',
    'allen',
    'almeida',
    'antonelli',
    'agnesi',
    'archimedes',
    'ardinghelli',
    'aryabhata',
    'austin',
    'babbage',
    'banach',
    'banzai',
    'bardeen',
    'bartik',
    'bassi',
    'beaver',
    'bell',
    'benz',
    'bhabha',
    'bhaskara',
    'black',
    'blackburn',
    'blackwell',
    'bohr',
    'booth',
    'borg',
    'bose',
    'bouman',
    'boyd',
    'brahmagupta',
    'brattain',
    'brown',
    'buck',
    'burnell',
    'cannon',
    'carson',
    'cartwright',
    'carver',
    'cerf',
    'chandrasekhar',
    'chaplygin',
    'chatelet',
    'chatterjee',
    'chebyshev',
    'cohen',
    'chaum',
    'clarke',
    'colden',
    'cori',
    'cray',
    'curran',
    'curie',
    'darwin',
    'davinci',
    'dewdney',
    'dhawan',
    'diffie',
    'dijkstra',
    'dirac',
    'driscoll',
    'dubinsky',
    'easley',
    'edison',
    'einstein',
    'elbakyan',
    'elgamal',
    'elion',
    'ellis',
    'engelbart',
    'euclid',
    'euler',
    'faraday',
    'feistel',
    'fermat',
    'fermi',
    'feynman',
    'franklin',
    'gagarin',
    'galileo',
    'galois',
    'ganguly',
    'gates',
    'gauss',
    'germain',
    'goldberg',
    'goldstine',
    'goldwasser',
    'golick',
    'goodall',
    'gould',
    'greider',
    'grothendieck',
    'haibt',
    'hamilton',
    'haslett',
    'hawking',
    'hellman',
    'heisenberg',
    'hermann',
    'herschel',
    'hertz',
    'heyrovsky',
    'hodgkin',
    'hofstadter',
    'hoover',
    'hopper',
    'hugle',
    'hypatia',
    'ishizaka',
    'jackson',
    'jang',
    'jemison',
    'jennings',
    'jepsen',
    'johnson',
    'joliot',
    'jones',
    'kalam',
    'kapitsa',
    'kare',
    'keldysh',
    'keller',
    'kepler',
    'khayyam',
    'khorana',
    'kilby',
    'kirch',
    'knuth',
    'kowalevski',
    'lalande',
    'lamarr',
    'lamport',
    'leakey',
    'leavitt',
    'lederberg',
    'lehmann',
    'lewin',
    'lichterman',
    'liskov',
    'lovelace',
    'lumiere',
    'mahavira',
    'margulis',
    'matsumoto',
    'maxwell',
    'mayer',
    'mccarthy',
    'mcclintock',
    'mclaren',
    'mclean',
    'mcnulty',
    'mendel',
    'mendeleev',
    'meitner',
    'meninsky',
    'merkle',
    'mestorf',
    'mirzakhani',
    'moore',
    'morse',
    'murdock',
    'moser',
    'napier',
    'nash',
    'neumann',
    'newton',
    'nightingale',
    'nobel',
    'noether',
    'northcutt',
    'noyce',
    'panini',
    'pare',
    'pascal',
    'pasteur',
    'payne',
    'perlman',
    'pike',
    'poincare',
    'poitras',
    'proskuriakova',
    'ptolemy',
    'raman',
    'ramanujan',
    'ride',
    'montalcini',
    'ritchie',
    'rhodes',
    'robinson',
    'roentgen',
    'rosalind',
    'rubin',
    'saha',
    'sammet',
    'sanderson',
    'satoshi',
    'shamir',
    'shannon',
    'shaw',
    'shirley',
    'shockley',
    'shtern',
    'sinoussi',
    'snyder',
    'solomon',
    'spence',
    'stonebraker',
    'sutherland',
    'swanson',
    'swartz',
    'swirles',
    'taussig',
    'tereshkova',
    'tesla',
    'tharp',
    'thompson',
    'torvalds',
    'tu',
    'turing',
    'varahamihira',
    'vaughan',
    'visvesvaraya',
    'volhard',
    'villani',
    'wescoff',
    'wilbur',
    'wiles',
    'williams',
    'williamson',
    'wilson',
    'wing',
    'wozniak',
    'wright',
    'wu',
    'yalow',
    'yonath',
    'zhukovsky',
]


def get_random_name() -> str:
    adjective = adjectives[random.randrange(0, len(adjectives))]
    surname = surnames[random.randrange(0, len(surnames))]
    return ''.join([adjective, '-', surname])



from typing import Any, Dict, Iterator, List, Optional

import pkg_resources
from requests.exceptions import ReadTimeout

from finetuner.client.base import _BaseClient
from finetuner.constants import (
    API_VERSION,
    CONFIG,
    CPUS,
    DELETE,
    DESCRIPTION,
    DEVICE,
    EXAMPLES,
    EXPERIMENTS,
    FINETUNER_VERSION,
    GET,
    GPUS,
    LOGS,
    LOGSTREAM,
    METRICS,
    NAME,
    POST,
    RUNS,
    STATUS,
)

_finetuner_core_version = pkg_resources.get_distribution('finetuner-stubs').version


class FinetunerV1Client(_BaseClient):
    """
    The Finetuner v1 API client.
    """

    """ Experiment API """

    def create_experiment(
        self, name: str = 'default', description: Optional[str] = ''
    ) -> Dict[str, Any]:
        """Create a new experiment.

        :param name: The name of the experiment.
        :param description: Optional description of the experiment.
        :return: Created experiment.
        """
        url = self._construct_url(self._base_url, API_VERSION, EXPERIMENTS) + '/'
        return self._handle_request(
            url=url, method=POST, json_data={NAME: name, DESCRIPTION: description}
        )

    def get_experiment(self, name: str) -> dict:
        """Get an experiment by its name.

        :param name: The name of the experiment.
        :return: Requested experiment.
        """
        url = self._construct_url(self._base_url, API_VERSION, EXPERIMENTS, name)
        return self._handle_request(url=url, method=GET)

    def list_experiments(self, page: int = 1, size: int = 50) -> Dict[str, Any]:
        """List every experiment.

        :param page: The page index.
        :param size: The number of experiments to retrieve.
        :return: Paginated results as a dict, where `items` are the `Experiment`s being
            retrieved.

        ..note:: `page` and `size` works together. For example, page 1 size 50 gives
            the 50 experiments in the first page. To get 50-100, set `page` as 2.
        ..note:: The maximum number for `size` per page is 100.
        """
        params = {'page': page, 'size': size}
        url = self._construct_url(self._base_url, API_VERSION, EXPERIMENTS) + '/'
        return self._handle_request(url=url, method=GET, params=params)

    def delete_experiment(self, name: str) -> Dict[str, Any]:
        """Delete an experiment given its name.

        :param name: The name of the experiment.
        :return: Experiment to be deleted.
        """
        url = self._construct_url(self._base_url, API_VERSION, EXPERIMENTS, name)
        return self._handle_request(url=url, method=DELETE)

    def delete_experiments(self) -> List[dict]:
        """Delete all experiments.

        :return: Experiments to be deleted.
        """
        url = self._construct_url(self._base_url, API_VERSION, EXPERIMENTS) + '/'
        return self._handle_request(url=url, method=DELETE)

    """ Run API """

    def get_run(self, experiment_name: str, run_name: str) -> dict:
        """Get a run by its name and experiment.

        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :return: Requested run.
        """
        url = self._construct_url(
            self._base_url, API_VERSION, EXPERIMENTS, experiment_name, RUNS, run_name
        )
        return self._handle_request(url=url, method=GET)

    def list_runs(
        self, experiment_name: Optional[str] = None, page: int = 50, size: int = 50
    ) -> Dict[str, Any]:
        """List all created runs inside a given experiment.

        If no experiment is specified, list runs for all available experiments.
        :param experiment_name: The name of the experiment.
        :param page: The page index.
        :param size: Number of runs to retrieve.
        :return: Paginated results as a dict, where `items` are the `Runs` being
            retrieved.

        ..note:: `page` and `size` works together. For example, page 1 size 50 gives
            the 50 runs in the first page. To get 50-100, set `page` as 2.
        ..note:: The maximum number for `size` per page is 100.
        """
        if not experiment_name:
            url = self._construct_url(self._base_url, API_VERSION, RUNS, RUNS)
        else:
            url = self._construct_url(
                self._base_url, API_VERSION, EXPERIMENTS, experiment_name, RUNS
            )
        params = {'page': page, 'size': size}
        return self._handle_request(url=url, method=GET, params=params)

    def delete_run(self, experiment_name: str, run_name: str) -> Dict[str, Any]:
        """Delete a run by its name and experiment.

        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :return: Deleted run.
        """
        url = self._construct_url(
            self._base_url, API_VERSION, EXPERIMENTS, experiment_name, RUNS, run_name
        )
        return self._handle_request(url=url, method=DELETE)

    def delete_runs(self, experiment_name: str) -> List[dict]:
        """Delete all runs inside a given experiment.

        :param experiment_name: The name of the experiment.
        :return: List of all deleted runs.
        """
        url = self._construct_url(
            self._base_url, API_VERSION, EXPERIMENTS, experiment_name, RUNS
        )
        return self._handle_request(url=url, method=DELETE)

    def get_run_status(self, experiment_name: str, run_name: str) -> dict:
        """Get a run status by its name and experiment.

        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :return: Run status.
        """
        url = self._construct_url(
            self._base_url,
            API_VERSION,
            EXPERIMENTS,
            experiment_name,
            RUNS,
            run_name,
            STATUS,
        )

        try:
            response = self._handle_request(url=url, method=GET, timeout=10)
        # In the event that the request times out, attempt the request one more time.
        except ReadTimeout:
            response = self._handle_request(url=url, method=GET, timeout=10)
        return response

    def get_run_logs(self, experiment_name: str, run_name: str) -> str:
        """Get a run logs by its name and experiment.

        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :return: Run logs.
        """
        url = self._construct_url(
            self._base_url,
            API_VERSION,
            EXPERIMENTS,
            experiment_name,
            RUNS,
            run_name,
            LOGS,
        )
        return self._handle_request(url=url, method=GET)

    def stream_run_logs(self, experiment_name: str, run_name: str) -> Iterator[str]:
        """Streaming log events to the client as ServerSentEvents.

        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :yield: A log entry.
        """
        url = self._construct_url(
            self._base_url,
            API_VERSION,
            EXPERIMENTS,
            experiment_name,
            RUNS,
            run_name,
            LOGSTREAM,
        )
        response = self._handle_request(url=url, method=GET, stream=True)
        for entry in response.iter_lines():
            if entry:
                decoded_message: str = entry.decode('utf-8', errors='ignore')
                sep_pos = decoded_message.find(': ')
                if sep_pos != -1:
                    msg_type, msg = (
                        decoded_message[:sep_pos],
                        decoded_message[sep_pos + 2 :],
                    )
                    if msg_type in ('data', 'event'):
                        yield msg

    def get_run_metrics(self, experiment_name: str, run_name: str) -> Dict[str, Any]:
        """Get evaluation metrics of a run by its name and experiment.
        This is only possible after the run has successfully finished.

        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :return: Dict of metrics before and after fine-tuning
        """
        url = self._construct_url(
            self._base_url,
            API_VERSION,
            EXPERIMENTS,
            experiment_name,
            RUNS,
            run_name,
            METRICS,
        )
        return self._handle_request(url=url, method=GET)

    def get_run_examples(self, experiment_name: str, run_name: str) -> Dict[str, Any]:
        """Get results of example queries of a run by the run's name and experiment.
        This is only possible after the run has successfully finished.

        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :return: Dict of results for example queries before and after fine-tuning.
        """
        url = self._construct_url(
            self._base_url,
            API_VERSION,
            EXPERIMENTS,
            experiment_name,
            RUNS,
            run_name,
            EXAMPLES,
        )
        return self._handle_request(url=url, method=GET)

    def create_run(
        self,
        experiment_name: str,
        run_name: str,
        run_config: dict,
        device: str,
        cpus: int,
        gpus: int,
    ) -> dict:
        """Create a run inside a given experiment.

        For optional parameters please visit our documentation (link).
        :param experiment_name: The name of the experiment.
        :param run_name: The name of the run.
        :param run_config: The run configuration.
        :param device: The device to use, either `cpu` or `gpu`.
        :param cpus: The number of CPUs to use.
        :param gpus: The number of GPUs to use.
        :return: Created run.
        """
        url = self._construct_url(
            self._base_url, API_VERSION, EXPERIMENTS, experiment_name, RUNS
        )
        return self._handle_request(
            url=url,
            method=POST,
            json_data={
                NAME: run_name,
                CONFIG: run_config,
                FINETUNER_VERSION: _finetuner_core_version,
                DEVICE: device,
                CPUS: cpus,
                GPUS: gpus,
            },
        )



DELETE = 'DELETE'
POST = 'POST'
GET = 'GET'
NAME = 'name'
HUBBLE_USER_ID = '_id'
ID = 'id'

HOST = 'JINA_FINETUNER_REGISTRY'
HUBBLE_REGISTRY = 'JINA_HUBBLE_REGISTRY'
DEFAULT_FINETUNER_HOST = 'https://api.compute.finetuner.fit'
DEFAULT_HUBBLE_REGISTRY = 'https://api.hubble.jina.ai'

CONFIG = 'config'
FINETUNER_VERSION = 'finetuner_version'
DEVICE = 'device'
CPUS = 'cpus'
GPUS = 'gpus'
NUM_WORKERS = 'num_workers'
RUNS = 'runs'
STATUS = 'status'
LOGS = 'logs'
LOGSTREAM = 'logstream'
METRICS = 'metrics'
EXAMPLES = 'examples'
EXPERIMENTS = 'experiments'
API_VERSION = 'api/v1'
AUTHORIZATION = 'Authorization'
CHARSET = 'Accept-Charset'
UTF_8 = 'utf-8'
TEXT = 'text'
TOKEN_PREFIX = 'token '
DATA = 'data'
TRAIN_DATA = 'train_data'
EVAL_DATA = 'eval_data'
VAL_SPLIT = 'val_split'
EVALUATE = 'evaluate'
ARTIFACTS_DIR = 'artifacts/'
MODEL = 'model'
MODEL_OPTIONS = 'model_options'
MODEL_ARTIFACT = 'model_artifact'
ARTIFACT = 'artifact'
ARTIFACT_ID = 'artifact_id'
DEFAULT_TAG_KEY = 'finetuner_label'
DEFAULT_TAG_SCORE_KEY = 'finetuner_score'
# Run status
CREATED = 'CREATED'
STARTED = 'STARTED'
FINISHED = 'FINISHED'
FAILED = 'FAILED'
DEFAULT_EXPERIMENT_NAME = 'default_experiment'
CREATED_AT = 'created_at'
DESCRIPTION = 'description'
FREEZE = 'freeze'
OUTPUT_DIM = 'output_dim'
MULTI_MODAL = 'multi_modal'
IMAGE_MODALITY = 'image_modality'
TEXT_MODALITY = 'text_modality'
HYPER_PARAMETERS = 'hyper_parameters'
LOSS = 'loss'
LOSS_OPTIONS = 'loss_options'
OPTIMIZER = 'optimizer'
LOSS_OPTIMIZER = 'loss_optimizer'
LOSS_OPTIMIZER_OPTIONS = 'loss_optimizer_options'
SAMPLER = 'sampler'
MINER = 'miner'
MINER_OPTIONS = 'miner_options'
BATCH_SIZE = 'batch_size'
LEARNING_RATE = 'learning_rate'
EPOCHS = 'epochs'
EXPERIMENT_NAME = 'experiment_name'
RUN_NAME = 'run_name'
OPTIMIZER_OPTIONS = 'optimizer_options'
SCHEDULER = 'scheduler'
SCHEDULER_OPTIONS = 'scheduler_options'
CALLBACKS = 'callbacks'
OPTIONS = 'options'
QUERY_DATA = 'query_data'
INDEX_DATA = 'index_data'
DA_PREFIX = 'finetuner-dastorage'
ONNX = 'to_onnx'
PUBLIC = 'public'
NUM_ITEMS_PER_CLASS = 'num_items_per_class'
VAL_SPLIT = 'val_split'



from typing import Any, Dict, List, Optional

from rich.console import Console
from rich.table import Table

from finetuner.model import list_model_classes

console = Console()


def print_model_table(model, task: Optional[str] = None):
    """Prints a table of model descriptions.

    :param model: Module with model definitions
    :param task: The fine-tuning task, should be one of `text-to-text`,
    """
    title = 'Finetuner backbones'
    if task:
        title += f': {task}'
    table = Table(title=title)
    header = model.get_header()
    model_display_names = set()

    for column in header:
        table.add_column(column, justify='right', style='cyan', no_wrap=False)

    for _, _model_class in list_model_classes().items():
        if _model_class.display_name not in model_display_names:
            row = model.get_row(_model_class)
            if task and row[1] != task:
                continue
            table.add_row(*row)
            model_display_names.add(_model_class.display_name)

    console.print(table)


def print_examples(stage: str, results: Dict[str, List[Any]], k: int = 5):
    """
    Prints a table of results of example queries from the evaluation data.

    :param stage: either 'before' or 'after'
    :param results: The example results to display
    :param k: maximal number of results per query to display
    """
    table = Table(
        title=f'Results {stage} fine-tuning:', show_header=False, title_justify='left'
    )
    table.add_column(justify='left', style='cyan', no_wrap=False)
    table.add_column(justify='left', style='cyan', no_wrap=False)
    for query in results:
        table.add_row('Query', str(query), style='yellow bold')
        for i, match in enumerate(results[query][:k]):
            table.add_row(f'Match {i+1}', str(match))
    console.print(table)


def print_metrics(stage: str, metrics: Dict[str, List[Any]]):
    """
    Prints a table of retrieval metrics.
    :param stage: either 'before' or 'after'
    :param metrics: dictionary with retrieval metrics before and after fine-tuning.
    """
    table = Table(title=f'Retrieval metrics {stage} fine-tuning:', title_justify='left')
    table.add_column('Retrieval Metric', justify='left', style='cyan', no_wrap=False)
    table.add_column('Value', justify='left', style='cyan', no_wrap=False)
    for metric, value in metrics.items():
        table.add_row(metric, str(value))
    console.print(table)



import os
from typing import List, Optional, Union

import requests

import hubble
from finetuner.client.session import _HeaderPreservingSession
from finetuner.constants import (
    AUTHORIZATION,
    CHARSET,
    DATA,
    HOST,
    HUBBLE_USER_ID,
    TEXT,
    TOKEN_PREFIX,
    UTF_8,
)
from finetuner.excepts import FinetunerServerError


class _BaseClient:
    """
    Base Finetuner API client.
    """

    def __init__(self):
        self._base_url = os.environ.get(HOST)
        self._session = self._get_client_session()
        self.hubble_client = hubble.Client(max_retries=None, jsonify=True)
        self.hubble_user_id = self._get_hubble_user_id()

    def _get_hubble_user_id(self):
        user_info = self.hubble_client.get_user_info()
        if user_info['code'] >= 400:
            # will implement error-handling later
            pass
        hubble_user_id = user_info[DATA][HUBBLE_USER_ID]
        return hubble_user_id

    @staticmethod
    def _get_client_session() -> _HeaderPreservingSession:
        session = _HeaderPreservingSession(trusted_domains=[])
        api_token = TOKEN_PREFIX + str(hubble.Auth.get_auth_token())
        session.headers.update({CHARSET: UTF_8, AUTHORIZATION: api_token})
        return session

    @staticmethod
    def _construct_url(*args) -> str:
        return '/'.join(args)

    def _handle_request(
        self,
        url: str,
        method: str,
        params: Optional[dict] = None,
        json_data: Optional[dict] = None,
        stream: bool = False,
        timeout: Optional[int] = None,
    ) -> Union[dict, List[dict], str, requests.Response]:
        """The base request handler.

        :param url: The url of the request.
        :param method: The request type (GET, POST or DELETE).
        :param params: Optional parameters for the request.
        :param json_data: Optional data payloads to be sent along with the request.
        :param stream: If the request is a streaming request set to True.
        :return: Response to the request.
        """
        response = self._session.request(
            url=url,
            method=method,
            json=json_data,
            params=params,
            allow_redirects=True,
            stream=stream,
            timeout=timeout,
        )
        if not response.ok:
            raise FinetunerServerError(
                message=response.reason,
                code=response.status_code,
                details=response.json()['detail'],
            )
        if stream:
            return response
        else:
            if TEXT in response.headers['content-type']:
                return response.text
            return response.json()



from typing import List

from requests import Session
from requests.utils import urlparse


class _HeaderPreservingSession(Session):
    def __init__(self, trusted_domains: List[str]):
        super(_HeaderPreservingSession, self).__init__()
        self._trusted_domains = trusted_domains

    def rebuild_auth(self, prepared_request, response):
        """
        Keep headers upon redirect as long as we are on any of the
        self._trusted_domains
        """
        headers = prepared_request.headers
        url = prepared_request.url
        if 'Authorization' in headers:
            _original_parsed = urlparse(response.request.url)
            _redirect_parsed = urlparse(url)
            _original_domain = '.'.join(_original_parsed.hostname.split('.')[-2:])
            _redirect_domain = '.'.join(_redirect_parsed.hostname.split('.')[-2:])
            if (
                _original_domain != _redirect_domain
                and _original_domain not in self._trusted_domains
                and _redirect_domain not in self._trusted_domains
            ):
                del headers['Authorization']



from _finetuner.runner.stubs import model
from _finetuner.runner.stubs.model import *  # noqa F401
from _finetuner.runner.stubs.model import _EmbeddingModelStub


def get_header() -> Tuple[str, ...]:
    """Get table header."""
    return 'name', 'task', 'output_dim', 'architecture', 'description'


def get_row(model_stub) -> Tuple[str, ...]:
    """Get table row."""
    return (
        model_stub.display_name,
        model_stub.task,
        str(model_stub.output_shape[1]),
        model_stub.architecture,
        model_stub.description,
    )


def list_model_classes() -> Dict[str, ModelStubType]:
    rv = {}
    members = inspect.getmembers(model, inspect.isclass)
    for name, stub in members:
        if (
            name != 'MLPStub'
            and not name.startswith('_')
            and type(stub) != type
            and issubclass(stub, _EmbeddingModelStub)
        ):
            rv[name] = stub
    return rv



from typing import Any, Dict, List, Optional, Union

from docarray import DocumentArray

import hubble
from finetuner.client import FinetunerV1Client
from finetuner.constants import CREATED_AT, DESCRIPTION, NAME, STATUS
from finetuner.data import CSVOptions
from finetuner.excepts import FinetunerServerError
from finetuner.experiment import Experiment
from finetuner.run import Run
from hubble import login_required


class Finetuner:
    """Finetuner class."""

    def __init__(self):
        self._client = None
        self._default_experiment = None
        self._default_experiment_name = 'default'

    def login(self, force: bool = False, interactive: Optional[bool] = None):
        """Login to Hubble account, initialize a client object
        and create a default experiment.

        :param force: If set to true, overwrite token and re-login.

        Note: Calling `login` is necessary for using finetuner.
        """
        hubble.login(
            force=force, post_success=self._init_state, interactive=interactive
        )

    @login_required
    def _init_state(self):
        """Initialize client and default experiment."""
        self._client = FinetunerV1Client()
        self._default_experiment = self._get_default_experiment()

    def _get_default_experiment(self) -> Experiment:
        """Create or retrieve (if it already exists) a default experiment
        for the current working directory."""
        for experiment in self.list_experiments():
            if experiment.name == self._default_experiment_name:
                return experiment
        return self.create_experiment(name=self._default_experiment_name)

    @login_required
    def create_experiment(self, name: str = 'default') -> Experiment:
        """Create an experiment.

        :param name: Optional name of the experiment. If `None`,
            the experiment is named after the current directory.
        :return: An `Experiment` object.
        """
        try:
            experiment = self._client.get_experiment(name=name)
        except FinetunerServerError:
            experiment = self._client.create_experiment(name=name)
        return Experiment(
            client=self._client,
            name=experiment[NAME],
            status=experiment[STATUS],
            created_at=experiment[CREATED_AT],
            description=experiment[DESCRIPTION],
        )

    @login_required
    def get_experiment(self, name: str) -> Experiment:
        """Get an experiment by its name.

        :param name: Name of the experiment.
        :return: An `Experiment` object.
        """
        experiment = self._client.get_experiment(name=name)
        return Experiment(
            client=self._client,
            name=experiment[NAME],
            status=experiment[STATUS],
            created_at=experiment[CREATED_AT],
            description=experiment[DESCRIPTION],
        )

    @login_required
    def list_experiments(self, page: int = 1, size: int = 50) -> List[Experiment]:
        """List every experiment.

        :param page: The page index.
        :param size: The number of experiments to retrieve.
        :return: A list of :class:`Experiment` instance.

        ..note:: `page` and `size` works together. For example, page 1 size 50 gives
            the 50 experiments in the first page. To get 50-100, set `page` as 2.
        ..note:: The maximum number for `size` per page is 100.
        """
        experiments = self._client.list_experiments(page=page, size=size)['items']

        return [
            Experiment(
                client=self._client,
                name=experiment[NAME],
                status=experiment[STATUS],
                created_at=experiment[CREATED_AT],
                description=experiment[DESCRIPTION],
            )
            for experiment in experiments
        ]

    @login_required
    def delete_experiment(self, name: str) -> Experiment:
        """Delete an experiment by its name.
        :param name: Name of the experiment.
        :return: Deleted experiment.
        """
        experiment = self._client.delete_experiment(name=name)
        return Experiment(
            client=self._client,
            name=experiment[NAME],
            status=experiment[STATUS],
            created_at=experiment[CREATED_AT],
            description=experiment[DESCRIPTION],
        )

    @login_required
    def delete_experiments(self) -> List[Experiment]:
        """Delete every experiment.
        :return: List of deleted experiments.
        """
        experiments = self._client.delete_experiments()
        return [
            Experiment(
                client=self._client,
                name=experiment[NAME],
                status=experiment[STATUS],
                created_at=experiment[CREATED_AT],
                description=experiment[DESCRIPTION],
            )
            for experiment in experiments
        ]

    @login_required
    def create_run(
        self,
        model: str,
        train_data: Union[str, DocumentArray],
        eval_data: Optional[Union[str, DocumentArray]] = None,
        val_split: float = 0.0,
        model_artifact: Optional[str] = None,
        run_name: Optional[str] = None,
        description: Optional[str] = None,
        experiment_name: Optional[str] = None,
        model_options: Optional[Dict[str, Any]] = None,
        loss: str = 'TripletMarginLoss',
        miner: Optional[str] = None,
        miner_options: Optional[Dict[str, Any]] = None,
        optimizer: str = 'Adam',
        optimizer_options: Optional[Dict[str, Any]] = None,
        learning_rate: Optional[float] = None,
        epochs: int = 5,
        batch_size: Optional[int] = None,
        callbacks: Optional[List[Any]] = None,
        scheduler: Optional[str] = None,
        scheduler_options: Optional[Dict[str, Any]] = None,
        freeze: bool = False,
        output_dim: Optional[int] = None,
        device: str = 'cuda',
        num_workers: int = 4,
        to_onnx: bool = False,
        csv_options: Optional[CSVOptions] = None,
        public: bool = False,
        num_items_per_class: int = 4,
        sampler: str = 'auto',
        loss_optimizer: Optional[str] = None,
        loss_optimizer_options: Optional[Dict[str, Any]] = None,
    ) -> Run:
        """Create a run.

        If an experiment name is not specified, the run will be created in the default
        experiment.

        :return: A `Run` object.
        """
        if not experiment_name:
            experiment = self._default_experiment
        else:
            experiment = self.get_experiment(name=experiment_name)
        if not experiment:
            raise ValueError(
                (
                    'Unable to start finetuning run as experiment is `None`. '
                    'Make sure you have logged in using `finetuner.login(force=True)`.'
                )
            )
        return experiment.create_run(
            model=model,
            train_data=train_data,
            eval_data=eval_data,
            val_split=val_split,
            model_artifact=model_artifact,
            run_name=run_name,
            description=description,
            model_options=model_options or {},
            loss=loss,
            miner=miner,
            miner_options=miner_options,
            optimizer=optimizer,
            optimizer_options=optimizer_options,
            learning_rate=learning_rate,
            epochs=epochs,
            batch_size=batch_size,
            callbacks=callbacks or [],
            scheduler=scheduler,
            scheduler_options=scheduler_options,
            freeze=freeze,
            output_dim=output_dim,
            device=device,
            num_workers=num_workers,
            to_onnx=to_onnx,
            csv_options=csv_options,
            public=public,
            num_items_per_class=num_items_per_class,
            sampler=sampler,
            loss_optimizer=loss_optimizer,
            loss_optimizer_options=loss_optimizer_options,
        )

    @login_required
    def get_run(self, run_name: str, experiment_name: Optional[str] = None) -> Run:
        """Get run by its name and (optional) experiment.

        If an experiment name is not specified, we'll look for the run in the default
        experiment.

        :param run_name: Name of the run.
        :param experiment_name: Optional name of the experiment.
        :return: A `Run` object.
        """
        if not experiment_name:
            experiment = self._default_experiment
        else:
            experiment = self.get_experiment(name=experiment_name)
        return experiment.get_run(name=run_name)

    @login_required
    def list_runs(
        self, experiment_name: Optional[str] = None, page: int = 1, size: int = 50
    ) -> List[Run]:
        """List all created runs inside a given experiment.

        If no experiment is specified, list runs for all available experiments.
        :param experiment_name: The name of the experiment.
        :param page: The page index.
        :param size: Number of runs to retrieve.
        :return: List of all runs.

        ..note:: `page` and `size` works together. For example, page 1 size 50 gives
            the 50 runs in the first page. To get 50-100, set `page` as 2.
        ..note:: The maximum number for `size` per page is 100.
        """
        if not experiment_name:
            experiments = self.list_experiments()
        else:
            experiments = [self.get_experiment(name=experiment_name)]
        runs = []
        for experiment in experiments:
            runs.extend(experiment.list_runs(page=page, size=size))
        return runs

    @login_required
    def delete_run(self, run_name: str, experiment_name: Optional[str] = None):
        """Delete a run.

        If an experiment name is not specified, we'll look for the run in the default
        experiment.

        :param run_name: Name of the run.
        :param experiment_name: Optional name of the experiment.
        """
        if not experiment_name:
            experiment = self._default_experiment
        else:
            experiment = self.get_experiment(name=experiment_name)
        experiment.delete_run(name=run_name)

    @login_required
    def delete_runs(self, experiment_name: Optional[str] = None):
        """Delete every run.

        If an experiment name is not specified, we'll delete every run across all
        experiments.

        :param experiment_name: Optional name of the experiment.
        """
        if not experiment_name:
            experiments = self.list_experiments()
        else:
            experiments = [self.get_experiment(name=experiment_name)]
        for experiment in experiments:
            experiment.delete_runs()

    @login_required
    def get_token(self) -> str:
        return hubble.Auth.get_auth_token()



from setuptools import find_packages, setup

# package name
_name = 'finetuner'

# package long description
try:
    with open('README.md', encoding='utf8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''


if __name__ == '__main__':
    setup(
        name=_name,
        packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),
        include_package_data=True,
        description='Task-oriented finetuning for better embeddings on neural search.',
        author='Jina AI',
        author_email='hello@jina.ai',
        url='https://github.com/jina-ai/finetuner/',
        license='Apache 2.0',
        download_url='https://github.com/jina-ai/finetuner/tags',
        long_description=_long_description,
        long_description_content_type='text/markdown',
        zip_safe=False,
        setup_requires=['setuptools>=18.0', 'wheel'],
        install_requires=[
            'docarray[common]>=0.21.0',
            'trimesh==3.16.4',
            'finetuner-stubs==0.13.3',
            'jina-hubble-sdk==0.33.1',
        ],
        extras_require={
            'full': [
                'finetuner-commons==0.13.3',
            ],
            'test': [
                'black==22.3.0',
                'flake8==5.0.4',
                'isort==5.10.1',
                'pytest==7.0.0',
                'pytest-cov==3.0.0',
                'pytest-mock==3.7.0',
            ],
        },
        python_requires='>=3.8.0',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'License :: OSI Approved :: Apache Software License',
            'Environment :: Console',
            'Operating System :: OS Independent',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
        project_urls={
            'Documentation': 'https://finetuner.jina.ai',
            'Source': 'https://github.com/jina-ai/finetuner/',
            'Tracker': 'https://github.com/jina-ai/finetuner/issues',
        },
        keywords=(
            'jina neural-search neural-network deep-learning pretraining '
            'fine-tuning pretrained-models triplet-loss metric-learning '
            'siamese-network few-shot-learning'
        ),
    )



from dataclasses import fields
from typing import Any, Dict, List, Optional, TextIO, Union

from _finetuner.runner.stubs import config
from docarray import DocumentArray

from finetuner.callback import EvaluationCallback
from finetuner.client import FinetunerV1Client
from finetuner.constants import (
    BATCH_SIZE,
    CALLBACKS,
    CONFIG,
    CREATED_AT,
    DESCRIPTION,
    DEVICE,
    EPOCHS,
    EVAL_DATA,
    FREEZE,
    LEARNING_RATE,
    LOSS,
    LOSS_OPTIMIZER,
    LOSS_OPTIMIZER_OPTIONS,
    MINER,
    MINER_OPTIONS,
    MODEL_ARTIFACT,
    MODEL_OPTIONS,
    NAME,
    NUM_ITEMS_PER_CLASS,
    NUM_WORKERS,
    ONNX,
    OPTIMIZER,
    OPTIMIZER_OPTIONS,
    OUTPUT_DIM,
    PUBLIC,
    SAMPLER,
    SCHEDULER,
    SCHEDULER_OPTIONS,
    VAL_SPLIT,
)
from finetuner.data import CSVContext
from finetuner.hubble import push_data
from finetuner.names import get_random_name
from finetuner.run import Run


class Experiment:
    """Class for an experiment.

    :param client: Client object for sending api requests.
    :param name: Name of the experiment.
    :param status: Status of the experiment.
    :param created_at: Creation time of the experiment.
    :param description: Optional description of the experiment.
    """

    def __init__(
        self,
        client: FinetunerV1Client,
        name: str,
        status: str,
        created_at: str,
        description: Optional[str] = '',
    ):
        self._client = client
        self._name = name
        self._status = status
        self._created_at = created_at
        self._description = description

    @property
    def name(self) -> str:
        """Get the name of the :class:`Experiment`."""
        return self._name

    @property
    def status(self) -> str:
        """Get the status of the :class:`Experiment`."""
        return self._status

    def get_run(self, name: str) -> Run:
        """Get a :class:`Run` given a `name`.

        :param name: Name of the run.
        :return: A `Run` object.
        """
        run = self._client.get_run(experiment_name=self._name, run_name=name)
        run = Run(
            name=run[NAME],
            config=run[CONFIG],
            created_at=run[CREATED_AT],
            description=run[DESCRIPTION],
            experiment_name=self._name,
            client=self._client,
        )
        return run

    def list_runs(self, page: int = 50, size: int = 50) -> List[Run]:
        """List all :class:`Run`.

        :param page: The page index.
        :param size: The number of runs to retrieve per page.
        :return: A list of :class:`Run` instance.

        ..note:: `page` and `size` works together. For example, page 1 size 50 gives
            the 50 runs in the first page. To get 50-100, set `page` as 2.
        ..note:: The maximum number for `size` per page is 100.
        """
        runs = self._client.list_runs(experiment_name=self._name, page=page, size=size)[
            'items'
        ]
        return [
            Run(
                name=run[NAME],
                config=run[CONFIG],
                created_at=run[CREATED_AT],
                description=run[DESCRIPTION],
                experiment_name=self._name,
                client=self._client,
            )
            for run in runs
        ]

    def delete_run(self, name: str):
        """Delete a :class:`Run` by its name.

        :param name: Name of the run.
        """
        self._client.delete_run(experiment_name=self._name, run_name=name)

    def delete_runs(self):
        """Delete all :class:`Run` inside the :class:`Experiment`."""
        self._client.delete_runs(experiment_name=self._name)

    def create_run(
        self,
        model: str,
        train_data: Union[str, TextIO, DocumentArray],
        run_name: Optional[str] = None,
        eval_data: Optional[Union[str, TextIO, DocumentArray]] = None,
        csv_options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Run:
        """Create a :class:`Run` inside the :class:`Experiment`."""
        if not run_name:
            run_name = get_random_name()

        eval_callback = None
        callbacks = kwargs[CALLBACKS] if kwargs.get(CALLBACKS) else []
        for callback in callbacks:
            if isinstance(callback, EvaluationCallback):
                eval_callback = callback

        csv_context = CSVContext(model=model, options=csv_options)
        train_data = csv_context.build_dataset(data=train_data)

        eval_data = csv_context.build_dataset(data=eval_data) if eval_data else None

        if eval_callback:
            eval_callback.query_data = csv_context.build_dataset(
                data=eval_callback.query_data,
            )
            eval_callback.index_data = csv_context.build_dataset(
                eval_callback.index_data
            )

        train_data, eval_data, query_data, index_data = push_data(
            experiment_name=self._name,
            run_name=run_name,
            train_data=train_data,
            eval_data=eval_data,
            query_data=eval_callback.query_data if eval_callback else None,
            index_data=eval_callback.index_data if eval_callback else None,
        )
        if query_data or index_data:
            eval_callback.query_data = query_data
            eval_callback.index_data = index_data

        kwargs[EVAL_DATA] = eval_data

        config = self._create_finetuning_config(
            model=model,
            train_data=train_data,
            experiment_name=self._name,
            run_name=run_name,
            **kwargs,
        )

        device = kwargs.get(DEVICE, 'cuda')
        if device == 'cuda':
            device = 'gpu'

        num_workers = kwargs.get(NUM_WORKERS, 4)
        run = self._client.create_run(
            run_name=run_name,
            experiment_name=self._name,
            run_config=config,
            device=device,
            cpus=num_workers,
            gpus=1,
        )
        run = Run(
            client=self._client,
            name=run[NAME],
            experiment_name=self._name,
            config=run[CONFIG],
            created_at=run[CREATED_AT],
            description=run[DESCRIPTION],
        )
        return run

    @staticmethod
    def _create_finetuning_config(
        model: str,
        train_data: str,
        experiment_name: str,
        run_name: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """Create config for a :class:`Run`.

        :param model: Name of the model to be fine-tuned.
        :param train_data: Either a `DocumentArray` for training data or a
            name of the `DocumentArray` that is pushed on Hubble.
        :param experiment_name: Name of the experiment.
        :param run_name: Name of the run.
        :param kwargs: Optional keyword arguments for the run config.
        :return: Run parameters wrapped up as a config dict.
        """
        callbacks = kwargs[CALLBACKS] if kwargs.get(CALLBACKS) else []
        callbacks = [
            config.CallbackConfig(
                name=callback.__class__.__name__,
                options={
                    field.name: getattr(callback, field.name)
                    for field in fields(callback)
                },
            )
            for callback in callbacks
        ]
        public = kwargs[PUBLIC] if kwargs.get(PUBLIC) else False
        model = config.ModelConfig(
            name=model if not kwargs.get(MODEL_ARTIFACT) else None,
            artifact=kwargs.get(MODEL_ARTIFACT),
            output_dim=kwargs.get(OUTPUT_DIM),
        )
        if kwargs.get(FREEZE):
            model.freeze = kwargs.get(FREEZE)
        if kwargs.get(MODEL_OPTIONS):
            model.options = kwargs.get(MODEL_OPTIONS)
        if kwargs.get(ONNX):
            model.to_onnx = kwargs.get(ONNX)

        data = config.DataConfig(
            train_data=train_data,
            eval_data=kwargs.get(EVAL_DATA),
            val_split=kwargs.get(VAL_SPLIT, 0.0),
            num_items_per_class=kwargs.get(NUM_ITEMS_PER_CLASS, 4),
        )
        if kwargs.get(NUM_WORKERS):
            data.num_workers = kwargs.get(NUM_WORKERS)
        if kwargs.get(SAMPLER):
            data.sampler = kwargs.get(SAMPLER)

        hyper_parameters = config.HyperParametersConfig(
            miner=kwargs.get(MINER),
            learning_rate=kwargs.get(LEARNING_RATE),
        )
        if kwargs.get(LOSS):
            hyper_parameters.loss = kwargs.get(LOSS)
        if kwargs.get(OPTIMIZER):
            hyper_parameters.optimizer = kwargs.get(OPTIMIZER)
        if kwargs.get(OPTIMIZER_OPTIONS):
            hyper_parameters.optimizer_options = kwargs.get(OPTIMIZER_OPTIONS)
        if kwargs.get(MINER_OPTIONS):
            hyper_parameters.miner_options = kwargs.get(MINER_OPTIONS)
        if kwargs.get(BATCH_SIZE):
            hyper_parameters.batch_size = kwargs.get(BATCH_SIZE)
        if kwargs.get(EPOCHS):
            epochs = kwargs.get(EPOCHS)
            hyper_parameters.epochs = epochs
        if kwargs.get(SCHEDULER):
            hyper_parameters.scheduler = kwargs.get(SCHEDULER)
        if kwargs.get(SCHEDULER_OPTIONS):
            hyper_parameters.scheduler_options = kwargs.get(SCHEDULER_OPTIONS)
        if kwargs.get(LOSS_OPTIMIZER):
            hyper_parameters.loss_optimizer = kwargs.get(LOSS_OPTIMIZER)
        if kwargs.get(LOSS_OPTIMIZER_OPTIONS):
            hyper_parameters.loss_optimizer_options = kwargs.get(LOSS_OPTIMIZER_OPTIONS)

        finetuning_config = config.FinetuningConfig(
            model=model,
            data=data,
            callbacks=callbacks,
            hyper_parameters=hyper_parameters,
            public=public,
            experiment_name=experiment_name,
            run_name=run_name,
        )

        return finetuning_config.dict()
